/* Model.cpp */

#include <stdio.h>
#include <vector>
#include <boost/mpi.hpp>
#include "AgentId.h"
#include "RepastProcess.h"
#include "Utilities.h"
#include "Properties.h"
#include "initialize_random.h"
#include "Point.h"
#include <iostream>
#include <fstream>
#include "Model.h"
#include <iomanip>
#include  <mpi.h>

AgentPackageProvider::AgentPackageProvider(repast::SharedContext<Agent>* agentPtr): agents(agentPtr){ }

void AgentPackageProvider::providePackage(Agent * agent, std::vector<AgentPackage>& out){
    repast::AgentId id = agent->getId();
    AgentPackage package(id.id(), id.startingRank(), id.agentType(), id.currentRank(), agent->getdis(), agent->getvel(), agent->getacc(),agent->getupdis(), agent->getupvel(), agent->getupacc());
    out.push_back(package);
}

void AgentPackageProvider::provideContent(repast::AgentRequest req, std::vector<AgentPackage>& out){
    std::vector<repast::AgentId> ids = req.requestedAgents();
    for(size_t i = 0; i < ids.size(); i++){
        providePackage(agents->getAgent(ids[i]), out);
    }
}


AgentPackageReceiver::AgentPackageReceiver(repast::SharedContext<Agent>* agentPtr): agents(agentPtr){}

Agent * AgentPackageReceiver::createAgent(AgentPackage package){
    repast::AgentId id(package.id, package.rank, package.type, package.currentRank);
    Agent* agent = new Agent(id, package.dis, package.vel, package.acc, package.updis, package.upvel, package.upacc);

    
    return agent;
}

void AgentPackageReceiver::updateAgent(AgentPackage package){
    repast::AgentId id(package.id, package.rank, package.type);
    Agent * agent = agents->getAgent(id);
    agent->set(package.currentRank, package.dis, package.vel, package.acc, package.updis, package.upvel, package.upacc);
}


Model::Model(std::string propsFile, int argc, char** argv, boost::mpi::communicator* comm): context(comm){
	
	props = new repast::Properties(propsFile, argc, argv, comm);
	stopAt = repast::strToInt(props->getProperty("stop.at"));
	countOfAgents = repast::strToInt(props->getProperty("count.of.agents"));
	PrNx = repast::strToInt(props->getProperty("ProcessorNumx"));
	PrNy = repast::strToInt(props->getProperty("ProcessorNumy"));
	PrNz = repast::strToInt(props->getProperty("ProcessorNumz"));
	buffer = repast::strToInt(props->getProperty("buffer"));
	timestep = repast::strToDouble(props->getProperty("timestep"));  
	rcut = repast::strToDouble(props->getProperty("rcut"));  
	initializeRandom(*props, comm);
	PrNtotal = PrNx*PrNy*PrNz;
	
	if(repast::RepastProcess::instance()->rank() == 0) props->writeToSVFile("./output/record.csv");
	
	/// keep this part
	provider = new AgentPackageProvider(&context);
	receiver = new AgentPackageReceiver(&context);
    
    repast::Point<double> origin(-50,-50,-50); // 3D stuff
    repast::Point<double> extent(100, 100, 100); // 3D stuff
    
    
    repast::GridDimensions gd(origin, extent);
    
    std::vector<int> processDims;
    processDims.push_back(PrNx);
    processDims.push_back(PrNy);
    processDims.push_back(PrNz); // 3D stuff
    
    continuousSpace = new mdr3d ::SharedContinuousSpace3D<Agent, repast::WrapAroundBorders, 
				repast::SimpleAdder<Agent> >("AgentcontinuousSpace", gd, processDims, buffer, comm);
    std::cout << "RANK " << repast::RepastProcess::instance()->rank() << " BOUNDS: " << continuousSpace->bounds().origin() << std::endl;
    
   	context.addProjection(continuousSpace);
   	
   	
    /// move init() to main.cpp
}

Model::~Model(){
	delete props;
	delete provider;
	delete receiver;

}

void Model::init(){
	
	int rank = repast::RepastProcess::instance()->rank();
	
		double xMin = continuousSpace->bounds().origin().getX();
        double xMax = xMin + continuousSpace->bounds().extents().getX();
        double yMin = continuousSpace->bounds().origin().getY();
        double yMax = yMin + continuousSpace->bounds().extents().getY();
        double zMin = continuousSpace->bounds().origin().getZ();
        double zMax = zMin + continuousSpace->bounds().extents().getZ();
	
		//each processor has one agent in the middle of box
        dis.x=xMin+1./2*(xMax-xMin);
		dis.y=yMin+1./2*(yMax-yMin);
		dis.z=zMin+1./2*(zMax-zMin);

		//constant velocity
		vel.x=10;
		vel.y=10;
		vel.z=10;
		
		acc.x=0;
		acc.y=0;
		acc.z=0;
		
		updis.x=0;
		updis.y=0;
		updis.z=0; 
		upvel.x=0;
		upvel.y=0;
		upvel.z=0;
		upacc.x=0;
		upacc.y=0;
		upacc.z=0;
		
		repast::AgentId id(1, rank, 0);
		id.currentRank(rank);
		Agent* agent = new Agent(id, dis, vel, acc, updis, upvel, upacc);
        
        repast::Point<double> initialLocation(dis.x,dis.y,dis.z); // 3D stuff
		
		context.addAgent(agent);
        continuousSpace->moveTo(id, initialLocation);

	
		repast::RepastProcess::instance()->synchronizeProjectionInfo<Agent, AgentPackage, AgentPackageProvider, AgentPackageReceiver>(context, *provider, *receiver, *receiver);  
		repast::RepastProcess::instance()->synchronizeAgentStatus<Agent, AgentPackage,  AgentPackageProvider, AgentPackageReceiver>(context, *provider, *receiver, *receiver);

	
}

/// request transfer of an agent from different process to local
//this part remains the same as tutorial demo 
/*
void Model::requestAgents(){
	int rank = repast::RepastProcess::instance()->rank();
	int worldSize= repast::RepastProcess::instance()->worldSize();
	repast::AgentRequest req(rank);
	for(int i = 0; i < worldSize; i++){                     // For each process
		if(i != rank){                                      // ... except this one
			std::vector<Agent*> agents;        
			context.selectAgents(5, agents);                // Choose 5 local agents randomly
			for(size_t j = 0; j < agents.size(); j++){
				repast::AgentId local = agents[j]->getId();          // Transform each local agent's id into a matching non-local one
				repast::AgentId other(local.id(), i, 0);
				other.currentRank(i);
				req.addRequest(other);                      // Add it to the agent request
			}
		}
	}
    repast::RepastProcess::instance()->requestAgents<Agent, AgentPackage, AgentPackageProvider, AgentPackageReceiver>(context, req, *provider, *receiver, *receiver);

}

void Model::cancelAgentRequests(){
	int rank = repast::RepastProcess::instance()->rank();
	if(rank == 0) std::cout << "CANCELING AGENT REQUESTS" << std::endl;
	repast::AgentRequest req(rank);
	
	repast::SharedContext<Agent>::const_state_aware_iterator non_local_agents_iter  = context.begin(repast::SharedContext<Agent>::NON_LOCAL);
	repast::SharedContext<Agent>::const_state_aware_iterator non_local_agents_end   = context.end(repast::SharedContext<Agent>::NON_LOCAL);
	while(non_local_agents_iter != non_local_agents_end){
		req.addCancellation((*non_local_agents_iter)->getId());
		non_local_agents_iter++;
	}
    repast::RepastProcess::instance()->requestAgents<Agent, AgentPackage, AgentPackageProvider, AgentPackageReceiver>(context, req, *provider, *receiver, *receiver);
	
	std::vector<repast::AgentId> cancellations = req.cancellations();
	std::vector<repast::AgentId>::iterator idToRemove = cancellations.begin();
	while(idToRemove != cancellations.end()){
		context.importedAgentRemoved(*idToRemove);
		idToRemove++;
	}
}


void Model::removeLocalAgents(){
	int rank = repast::RepastProcess::instance()->rank();
	if(rank == 0) std::cout << "REMOVING LOCAL AGENTS" << std::endl;
	for(int i = 0; i < 5; i++){
		repast::AgentId id(i, rank, 0);
		repast::RepastProcess::instance()->agentRemoved(id);
		context.removeAgent(id);
	}
	repast::RepastProcess::instance()->synchronizeAgentStatus<Agent, AgentPackage, AgentPackageProvider, AgentPackageReceiver>(context, *provider, *receiver, *receiver);
}
 
void Model::connectAgentNetwork(){
	repast::SharedContext<Agent>::const_local_iterator iter    = context.localBegin();
	repast::SharedContext<Agent>::const_local_iterator iterEnd = context.localEnd();
	while(iter != iterEnd) {
		Agent* ego = &**iter;
		std::vector<Agent*> agents;
		agents.push_back(ego);                          // Omit self
		context.selectAgents(countOfAgents, agents, true);          // Choose other agents randomly and remove elements original in the agents from context
		// Make an undirected connection
		for(size_t i = 0; i < agents.size(); i++){
         	    std::cout << "CONNECTING: " << ego->getId() << " to " << agents[i]->getId() << std::endl;
  	  	    agentNetwork->addEdge(ego, agents[i]);	
		}
		iter++;
	}	
}



*/



void Model::doSomething(){
	int whichRank = 0;
	if(repast::RepastProcess::instance()->rank() == whichRank) std::cout << " TICK " << repast::RepastProcess::instance()->getScheduleRunner().currentTick() << std::endl;


	if(repast::RepastProcess::instance()->rank() == whichRank)
	{
		std::cout<< "is done" << std::endl;

	}

	     
	//initialize force to 0;
	std::vector<Agent*> agents;
	context.selectAgents(repast::SharedContext<Agent>::LOCAL, agents);
	std::vector<Agent*>::iterator it = agents.begin();
	while(it != agents.end()){
		(*it)->initforce();
		it++;
    }
    

		
}


void Model::Interaction(){
	
	repast::SharedContext<Agent>::const_local_iterator iter    = context.localBegin();
	repast::SharedContext<Agent>::const_local_iterator iterEnd = context.localEnd();

	
	while( iter != iterEnd) {
			//leapfrog step 1
            agentLoci.clear();
            continuousSpace->getLocation((*iter)->getId(), agentLoci);
            agentVeli=(*iter)->getvel();
            agentAcci=(*iter)->getacc();
            agentLoci[0]=agentLoci[0]+timestep*(agentVeli.x+timestep/2*agentAcci.x);
			agentLoci[1]=agentLoci[1]+timestep*(agentVeli.y+timestep/2*agentAcci.y);
			agentLoci[2]=agentLoci[2]+timestep*(agentVeli.z+timestep/2*agentAcci.z);
			Loci.x=agentLoci[0];
			Loci.y=agentLoci[1];
			Loci.z=agentLoci[2];
			(*iter)->setupdis(Loci);
				  
				  
			std::set<Agent*> agentsToPlay;   
            context.selectAgents(context.size(), agentsToPlay);	     
		    std::set<Agent*>::iterator agentToPlay = agentsToPlay.begin();    
            while(agentToPlay != agentsToPlay.end())  
            {
  
                  agentLocj.clear();               
				  continuousSpace->getLocation((*agentToPlay)->getId(), agentLocj);                  
                  agentVelj=(*agentToPlay)->getvel();                
                  agentAccj=(*agentToPlay)->getacc();			 
				  
				  agentLocj[0]=agentLocj[0]+timestep*(agentVelj.x+timestep/2*agentAccj.x);
				  agentLocj[1]=agentLocj[1]+timestep*(agentVelj.y+timestep/2*agentAccj.y);
				  agentLocj[2]=agentLocj[2]+timestep*(agentVelj.z+timestep/2*agentAccj.z);
				
				  dxij=agentLoci[0]-agentLocj[0];
				  dyij=agentLoci[1]-agentLocj[1];
				  dzij=agentLoci[2]-agentLocj[2];
					
				  dxy=sqrt(dxij*dxij+dyij*dyij+dzij*dzij);

				//use Lennard Jones potential as the interaction between agents
				if(dxy<=rcut && dxy!=0)
				{
					 redxy=1./dxy;
					 Fij=48*(pow(redxy,14.)-0.5*pow(redxy,8.));
					
					
					force.x=Fij*dxij;	
					force.y=Fij*dyij;
					force.z=Fij*dzij;
					(*iter)->addforce(force);
				
				
				}
				

			 agentToPlay++;
		    }

	    
		iter++;
	}
	
	std::set<Agent*> agentsToStore;   
    context.selectAgents(context.size(), agentsToStore);	
    std::set<Agent*>::iterator agentToAcross = agentsToStore.begin();
	int rank = repast::RepastProcess::instance()->rank();
	
		    
    while(agentToAcross != agentsToStore.end())
            {
		//leapfrog step 2		
		(*agentToAcross)->computupacc();
		agentAcci=(*agentToAcross)->getupacc();      //t+h
		agentAccj=(*agentToAcross)->getacc();         //t
		agentVelj=(*agentToAcross)->getvel();         //t
 		agentVeli.x=agentVelj.x+timestep/2*agentAccj.x+timestep/2*agentAcci.x;          //t+h
		agentVeli.y=agentVelj.y+timestep/2*agentAccj.y+timestep/2*agentAcci.y;
		agentVeli.z=agentVelj.z+timestep/2*agentAccj.z+timestep/2*agentAcci.z;
		
		(*agentToAcross)->setupvel(agentVeli);
		
		repast::AgentId id = (*agentToAcross)->getId();  

		
		if(id.currentRank()==rank)                // move local agents by displacement
		{
		(*agentToAcross)->setR();
		(*agentToAcross)->move(continuousSpace);
			}
				
				agentToAcross++;
				}
				
	continuousSpace->balance();
	
	repast::RepastProcess::instance()->synchronizeAgentStatus<Agent, AgentPackage,  AgentPackageProvider, AgentPackageReceiver>(context, *provider, *receiver, *receiver);
	repast::RepastProcess::instance()->synchronizeProjectionInfo<Agent, AgentPackage, AgentPackageProvider, AgentPackageReceiver>(context, *provider, *receiver, *receiver);  	
	
	////////////update dis vel and acc of local agents after moving
	repast::SharedContext<Agent>::const_local_iterator  lp1   = context.localBegin(); 
	repast::SharedContext<Agent>::const_local_iterator lpEnd1 = context.localEnd();
	while( lp1 != lpEnd1) {
		std::vector<double> agentLoc;
        continuousSpace->getLocation((*lp1)->getId(), agentLoc);
        Loci.x=agentLoc[0];
        Loci.y=agentLoc[1];
        Loci.z=agentLoc[2];
        (*lp1)->setupdis(Loci);
             
        
		(*lp1)->update();
		lp1++;	
                           }
	

    repast::RepastProcess::instance()->synchronizeAgentStatus<Agent, AgentPackage,  AgentPackageProvider, AgentPackageReceiver>(context, *provider, *receiver, *receiver);
    repast::RepastProcess::instance()->synchronizeProjectionInfo<Agent, AgentPackage, AgentPackageProvider, AgentPackageReceiver>(context, *provider, *receiver, *receiver);  

    
	}
	


void Model::initSchedule(repast::ScheduleRunner& runner){
	//~ runner.scheduleEvent(1, repast::Schedule::FunctorPtr(new repast::MethodFunctor<Model> (this, &Model::requestAgents)));
	//~ runner.scheduleEvent(1.1, repast::Schedule::FunctorPtr(new repast::MethodFunctor (this, &Model::connectAgentNetwork)));
	runner.scheduleEvent(1, 1, repast::Schedule::FunctorPtr(new repast::MethodFunctor<Model> (this, &Model::doSomething)));
	runner.scheduleEvent(1.1, 50, repast::Schedule::FunctorPtr(new repast::MethodFunctor<Model> (this, &Model::writeCFG)));
	runner.scheduleEvent(1.2, 50, repast::Schedule::FunctorPtr(new repast::MethodFunctor<Model> (this, &Model::writeothers)));
    runner.scheduleEvent(1.3, 1, repast::Schedule::FunctorPtr(new repast::MethodFunctor<Model> (this, &Model::Interaction)));
	//~ runner.scheduleEvent(4, repast::Schedule::FunctorPtr(new repast::MethodFunctor<Model> (this, &Model::removeLocalAgents)));
	runner.scheduleEndEvent(repast::Schedule::FunctorPtr(new repast::MethodFunctor<Model> (this, &Model::recordResults)));
	runner.scheduleStop(stopAt);
}

void Model::recordResults(){
	if(repast::RepastProcess::instance()->rank() == 0){
		props->putProperty("Result","Passed");
		std::vector<std::string> keyOrder;
		keyOrder.push_back("RunNumber");
		keyOrder.push_back("stop.at");
		keyOrder.push_back("Result");
		props->writeToSVFile("./output/results.csv", keyOrder);
    }
}

void Model::writeCFG(){
	
	int rank = repast::RepastProcess::instance()->rank();
	int tick=repast::RepastProcess::instance()->getScheduleRunner().currentTick();
	char *fn_tmp_config;
	char *buf = new char[20];
	strcpy(fn_tmp_config, "Cfg/s.");
	sprintf (buf, "%s%d",fn_tmp_config, tick); 
     
     
    double H[3][3];
    H[0][0] = 100.;
    H[0][1] = 0.;
    H[0][2] = 0.;
    H[1][0] = 0.;
    H[1][1] = 100.;
    H[1][2] = 0.;
    H[2][0] = 0.;
    H[2][1] = 0.;
    H[2][2] = 100.;

	
	
	const char *out_file_name = buf;
	std::ofstream out;

	
	if(rank == 0){

	    out.open(out_file_name);	

	out << "Number of particles = 8\n";
	out << "A = 3.33 Angstrom (basic length-scale)\n";
	out << "H0(1,1) = " << H[0][0] << " A\n";
	out << "H0(1,2) = " << H[0][1] << " A\n";
	out << "H0(1,3) = " << H[0][2] << " A\n";
	out << "H0(2,1) = " << H[1][0] << " A\n";
	out << "H0(2,2) = " << H[1][1] << " A\n";
	out << "H0(2,3) = " << H[1][2] << " A\n";
	out << "H0(3,1) = " << H[2][0] << " A\n";
	out << "H0(3,2) = " << H[2][1] << " A\n";
	out << "H0(3,3) = " << H[2][2] << " A\n";
	out << "entry_count = 7\n";
	out << "1\n";
	out << "Ar\n";

         out.close();
         
         
         out.open(out_file_name,std::ios::app); 
               std::vector<Agent*> agents;
	context.selectAgents(repast::SharedContext<Agent>::LOCAL, agents);
	std::vector<Agent*>::iterator it = agents.begin();
	while(it != agents.end()){
		
		dis=(*it)->getdis();
		vel=(*it)->getvel();
		
		acc=(*it)->getacc();
		mass=(*it)->getmass();
		out << (dis.x+50)/100 <<" "<< (dis.y+50)/100 <<" "<< (dis.z+50)/100<<" "<< vel.x <<" "<< vel.y <<" "<< vel.z << " " <<(*it)->getId().id()<< " " <<(*it)->getId().startingRank()<< " " <<(*it)->getId().agentType()<< " " <<(*it)->getId().currentRank() <<"\n";
		it++; 
         }
	 }
    
         out.close();
    delete buf;
    
}


void Model::writeothers(){
	
	
	
	int rank = repast::RepastProcess::instance()->rank();
	int tick=repast::RepastProcess::instance()->getScheduleRunner().currentTick();
	char *buf= new char[20];
	sprintf (buf, "Cfg/s.%d", tick);

	const char *out_file_name=buf;
	std::ofstream out;
    
    
     boost::mpi::communicator world;
     
    for(int i=1; i<PrNtotal; i++)
    {
		if(rank == i)
		{

	out.open(out_file_name,std::ios::app);
     
    std::vector<Agent*> agents;
	context.selectAgents(repast::SharedContext<Agent>::LOCAL, agents);
	std::vector<Agent*>::iterator it = agents.begin();
	while(it != agents.end()){
		
		dis=(*it)->getdis();
		vel=(*it)->getvel();
		
		acc=(*it)->getacc();
		out << (dis.x+50)/100 <<" "<< (dis.y+50)/100 <<" "<< (dis.z+50)/100<<" "<< vel.x <<" "<< vel.y <<" "<< vel.z << " " <<(*it)->getId().id()<< " " <<(*it)->getId().startingRank()<< " " <<(*it)->getId().agentType()<< " " <<(*it)->getId().currentRank() <<"\n";
		
		it++; 
                              }
         
         
         
	     }
	     out.close();
	  world.barrier();
    }
     
     
     delete []buf;
}

